package net.locusworks.lsproject.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.*;

import net.locusworks.lsproject.gui.helper.ImageHelper;
import net.locusworks.lsproject.gui.helper.RouterPath;
import net.locusworks.lsproject.object.Connection;
import net.locusworks.lsproject.object.Router;
import net.locusworks.lsproject.object.Network;
import net.locusworks.lsproject.util.Utility;


/**
 * Display a network.
 */
public class RouterViewerWindow extends JFrame implements MouseListener, 
														MouseMotionListener, 
														KeyListener {
	// Makes compiler happy.
	private static final long serialVersionUID = 5141131466102432149L;
	
	private JPanel surface;
	private Image routerImage;
	private int imageWidth;
	private int imageHeight;
	
	private String highlightedRouter;
	
	private Image offscreenImage;
	private Graphics buffer;
	
	private boolean[] keys;
	
	private Hashtable<String, RoutingTableWindow> routingTables = new Hashtable<String, RoutingTableWindow>();
	
	// This is a hack for a weird java-related bug where it does not draw pictures on the first paint.
	private boolean firstPaint = true;
	
	/**
	 * Construct a new RouterViewerWindow object.
	 */
	public RouterViewerWindow() {
		super("Network Viewer - " + Network.getInstance().getName());
		
		initialize();
	}
	
	/**
	 * Initialize the GUI components of this class.
	 */
	private void initialize() {
		routerImage = ImageHelper.getImageFromResource("assets/router.png");
		imageWidth = 129;
		imageHeight = 71;
		
		setIconImage(ImageHelper.getImageFromResource("assets/router_ico.png"));
		
		surface = new JPanel(true);
		surface.setSize(800, 600);
		surface.setBackground(Color.GRAY);
		
		setContentPane(surface);
		
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
		
		setSize(800, 600);
		
		setResizable(false);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setLocationRelativeTo(null);
		setVisible(true);
		
		// Track key states. Initialize all to false (not pressed).
		keys = new boolean[525];
		for (int i = 0; i < keys.length; i++) {
			keys[i] = false;
		}
		
		// Create the buffer.
		offscreenImage = surface.createImage(getWidth(), getHeight());
		buffer = offscreenImage.getGraphics();
		
		// Repaint the window every once in a while.
		new Thread() {
			public void run() {
				while (isVisible()) {
					try {
						Thread.sleep(500);
					}
					catch (InterruptedException e) {
						e.printStackTrace();
						return;
					}
					
					repaint();
				}
			}
		}.start();
	}
	
	/**
	 * Paint the routers to the screen.
	 */
	@Override 
	public void paint(Graphics g) {
		buffer.setColor(Color.GRAY);
		buffer.fillRect(0, 0, getWidth(), getHeight());
		
		buffer.setColor(Color.BLACK);
		
		// Draw connections.
		for (Router r : Network.getInstance().getRouters()) {
			for (Connection connection : r.getConnections()) {
				Router r1 = Network.getInstance().getRouterByIp(connection.getFirstParticipantAddress());
				Router r2 = Network.getInstance().getRouterByIp(connection.getSecondParticipantAddress());
				
				buffer.drawLine(r1.getPosition().x, r1.getPosition().y, 
					r2.getPosition().x, r2.getPosition().y);
				
				buffer.drawString(Integer.toString(connection.getCost()), 
					(r1.getPosition().x + r2.getPosition().x) / 2, 
					(r1.getPosition().y + r2.getPosition().y - 10) / 2);
			}
		}
		
		for (Router r : Network.getInstance().getRouters()) {
			Point position = r.getPosition();
			
			buffer.setColor(Color.BLACK);
			
			// Draw router IP addresses.
			buffer.drawString(r.getIpAddress(), position.x - (imageWidth / 2) + 5, 
				position.y - (imageHeight / 2) - 10);
			
			buffer.drawImage(routerImage, position.x - (imageWidth / 2), 
				position.y - (imageHeight / 2), imageWidth, imageHeight, null);
			
			// Highlight a different color if shift is pressed down.
			if (keys[KeyEvent.VK_SHIFT]) {
				buffer.setColor(Color.RED);
			}
			
			if (highlightedRouter != null && !highlightedRouter.equals("")) {
				if (r.getIpAddress().equals(highlightedRouter)) {
					buffer.drawRect(position.x - (imageWidth / 2), position.y - (imageHeight / 2),
						imageWidth, imageHeight);
				}
			}
		}
		
		illuminatePath(buffer);
		
		// Draw the buffer image to the screen.
		g.drawImage(offscreenImage, 0, 0, getWidth(), getHeight(), Color.GRAY, null);
		
		// Hack to attempt to get the first image to draw.
		if (firstPaint) {
			firstPaint = false;
			repaint();
		}
	}
	
	/**
	 * Highlight the path between a start and end router, if necessary.
	 * @param g Graphics device.
	 */
	public void illuminatePath(Graphics g) {
		// Draw highlighted router boxes.
		if (RouterPath.firstRouter != null) {
			final Point position = RouterPath.firstRouter.getPosition();
			buffer.setColor(new Color(0, 255, 255));
			buffer.drawRect(position.x - (imageWidth / 2), position.y - (imageHeight / 2),
					imageWidth, imageHeight);
		}
		
		if (RouterPath.secondRouter != null) {
			final Point position = RouterPath.secondRouter.getPosition();
			buffer.setColor(new Color(0, 255, 255));
			buffer.drawRect(position.x - (imageWidth / 2), position.y - (imageHeight / 2),
					imageWidth, imageHeight);
		}
		
		if (RouterPath.highlightedPath != null) {
			Router lastRouter = RouterPath.firstRouter;
			while (RouterPath.highlightedPath.hasNext()) {
				final Router next = RouterPath.highlightedPath.getNext();
				
				final Point start = lastRouter.getPosition();
				final Point end = next.getPosition();
				buffer.drawLine(start.x, start.y, end.x, end.y);
				
				lastRouter = next;
			}
			
			RouterPath.highlightedPath.reset();
		}
	}
	
	/**
	 * Overridden to prevent the frame from erasing the background.
	 */
	@Override public void update(Graphics g) {
		paint(g);
	}
	
	/**
	 * Spawn a new routing table window.
	 * @param r Router to display.
	 */
	public void spawnRoutingTable(Router r) {
		if (routingTables.contains(r.getIpAddress())) {
			RoutingTableWindow window = routingTables.get(r.getIpAddress());
			window.dispose();
			routingTables.remove(r.getIpAddress());
		}
		
		routingTables.put(r.getIpAddress(), new RoutingTableWindow(r));
	}
	
	public void spawnCostViwerWindow(Router r){
		CostViewerWindow window = new CostViewerWindow(r);
		window.createGUI();
	}
	
	/**
	 * Force each router to perform the algorithm.
	 */
	public void updateAllRouters() {
		for (Router router : Network.getInstance().getRouters()) {
			router.getRoutingTable().execute(router, null);
		}
	}

	/*
	 * Event handlers.
	 */
	/**
	 * Check if the mouse is over a router on the screen.
	 */
	private Router mouseIsOverRouter(MouseEvent e) {
		for (Router r : Network.getInstance().getRouters()) {
			if (Math.abs(r.getPosition().x - e.getPoint().x) < imageWidth / 2 &&
				Math.abs(r.getPosition().y - e.getPoint().y) < imageHeight / 2) {
				
				return r;
			}
		}
		
		return null;
	}
	
	@Override public void mouseClicked(MouseEvent e) {}
	@Override public void mouseEntered(MouseEvent e) {}
	@Override public void mouseExited(MouseEvent e) {}
	@Override public void mouseReleased(MouseEvent e) {}
	
	/**
	 * Check if the mouse was pressed down. This will highlight or unhighlight routers.
	 */
	@Override public void mousePressed(MouseEvent e) {
		// Collision detection.
		Router r = null;
		if ((r = mouseIsOverRouter(e)) != null) {
			// The mouse is over a router.
			if (highlightedRouter != null && !highlightedRouter.equals(r.getIpAddress())) {
				// The selected router is not currently highlighted.
				if (keys[KeyEvent.VK_SHIFT]) {
					// Shift is pressed down. User is trying to make a connection.
					Random rand = new Random();
					Connection connection = new Connection(
						highlightedRouter, r.getIpAddress(), rand.nextInt(5) + 1);
					
					if (r.getConnectionByParticipant(highlightedRouter, r.getIpAddress()) != null) {
						// Connection exists.
						return;
					}
					
					r.addConnection(connection);
					Network.getInstance().getRouterByIp(highlightedRouter).addConnection(connection);
					
					System.out.println("Added connection between " + highlightedRouter + " and " +
						r.getIpAddress() + ". Cost: " + connection.getCost() + ".");
					
					repaint();
					
					return;
				}
			}
			
			if (keys[KeyEvent.VK_CONTROL]) {
				// CTRL is pressed down -- select routers for highlighting.
				RouterPath.highlightOne(r);
				
				repaint();
				
				return;
			}
			
			highlightedRouter = r.getIpAddress();
			
			repaint();
			
			return;
		}
		
		if (keys[KeyEvent.VK_SHIFT] || keys[KeyEvent.VK_CONTROL]) {
			// Do not create a router if shift or control are held down.
			e.consume();
			return;
		}
		
		// If the program reaches this point, it will create a new router.
		
		int routerId = Network.getInstance().getRouters().size();
		String routerName = new String(new char[] { (char)(routerId + 65) });
		while (Network.getInstance().getRouterByIp(routerName) != null) {
			routerId++;
		}
		
		Network.getInstance().addRouter(new Router(routerName, e.getPoint()));
		highlightedRouter = routerName;
		
		System.out.println("Added new router: " + routerName);
		
		repaint();
	}
	
	/**
	 * Handle an event when the mouse is dragged.
	 */
	@Override public void mouseDragged(MouseEvent e) {
		// Check if a mouse is trying to drag a router across the screen.
		if (highlightedRouter != null && !highlightedRouter.equals("") 
				&& !keys[KeyEvent.VK_SHIFT] && !keys[KeyEvent.VK_CONTROL]) {
			Router r = Network.getInstance().getRouterByIp(highlightedRouter);
			if (r == null) {
				return;
			}
			
			Point offset = new Point();
			offset.x = r.getPosition().x + (e.getPoint().x - r.getPosition().x);
			offset.y = r.getPosition().y + (e.getPoint().y - r.getPosition().y);
			
			r.setPosition(offset);
			
			repaint();
		}
	}

	@Override public void mouseMoved(MouseEvent e) {}
	@Override public void keyTyped(KeyEvent e) {}

	@Override public void keyPressed(KeyEvent e) {
		keys[e.getKeyCode()] = true;
		
		if (highlightedRouter != null && 
				!highlightedRouter.equals("") && 
				keys[KeyEvent.VK_DELETE]) {
			Network.getInstance().removeRouterByIp(highlightedRouter);
			
			try {
				highlightedRouter = Network.getInstance().getRouters().get(0).toString();
			}
			catch (Exception ex) {}
			
			RouterPath.clear();
			
			updateAllRouters();
		}
		
		else if (keys[KeyEvent.VK_S] && ((e.getModifiers() & KeyEvent.VK_CONTROL) == 0)) {
			try {
				Utility.saveNetwork();
				
				System.out.println("Network saved to " + Network.getInstance().getName() + Utility.FILE_EXTENSION);
			}
			catch (IOException ex) {
				System.err.println("Couldn't save network. Details:");
				ex.printStackTrace();
			}
		}
		
		// Spawn routing table.
		else if (e.getKeyCode() == KeyEvent.VK_R) {
			spawnRoutingTable(Network.getInstance().getRouterByIp(highlightedRouter));
		}
		
		if (e.getKeyCode() == KeyEvent.VK_C){
			SwingUtilities.invokeLater(new Runnable(){
				public void run(){
					spawnCostViwerWindow(Network.getInstance().getRouterByIp(highlightedRouter));
				}
			});
		}
		
		repaint();
	}

	@Override public void keyReleased(KeyEvent e) {
		keys[e.getKeyCode()] = false;
		
		repaint();
	}
}
