package net.locusworks.lsproject.object;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

/**
 * Network class is capable of performing LS calculations and has the router information
 * 
 * 
 */
public class Network implements Serializable {
	// Makes compiler happy.
	private static final long serialVersionUID = -3073110726594813517L;
	
	private String name = null;
	private ArrayList<Router> routers = new ArrayList<Router>();
	
	private static Network instance;
	
	
	/**
	 * Create a new Network.
	 * @param _name Name of the network.
	 */
	private Network(String _name) {
		setName(_name);
	}
	
	/**
	 * Obtain the network's name.
	 * @return String containing the network's name.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Set a new name for the network, which is used as the file name for the network.
	 * @param _name Name of the network.
	 */
	public void setName(String _name) {
		name = _name;
	}
	
	/**
	 * Add a router to the network.
	 * @param r Router to add.
	 */
	public synchronized void addRouter(Router r) {
		routers.add(r);
	}
	
	/**
	 * Get the routers in this network.
	 * @return Array list of routers.
	 */
	public synchronized ArrayList<Router> getRouters() {
		return routers;
	}
	
	/**
	 * Get a router by it's IP address.
	 * @param ip IP address to search for.
	 * @return The router that is being searched for. Returns null if the router doesn't exist.
	 */
	public synchronized Router getRouterByIp(String ip) {
		for (Router r : routers) {
			if (r.getIpAddress().equals(ip)) {
				return r;
			}
		}
		
		return null;
	}
	
	public synchronized List<Router> getNeighboringRouters(Router _router){
		List<Router> neighboringRouters = new ArrayList<Router>(_router.getConnections().size());
		
		final String sourceAddr = _router.getIpAddress();
		for (Connection connection : _router.getConnections()){
			if (!connection.getFirstParticipantAddress().equals(sourceAddr)){
				neighboringRouters.add(getRouterByIp(connection.getFirstParticipantAddress()));
			}
			else{ 
				neighboringRouters.add(getRouterByIp(connection.getSecondParticipantAddress()));
			}
		}
		
		return neighboringRouters;
	}
	
	/**
	 * Remove a router from the network.
	 * @param ip IP address of the router.
	 */
	public synchronized void removeRouterByIp(String ip) {
		for (int i = 0; i < routers.size(); i++) {
			if (routers.get(i).getIpAddress().equals(ip)) {
				for (Connection c : routers.get(i).getConnections()) {
					final String otherRouterAddr = (!c.getFirstParticipantAddress().equals(ip) ?
						c.getFirstParticipantAddress() : c.getSecondParticipantAddress());
					
					Router r = getRouterByIp(otherRouterAddr);
					r.removeConnectionByParticipant(ip, otherRouterAddr);
				}
				
				routers.remove(i);
				
				break;
			}
		}
	}
	
	public static void createInstance(String networkName) {
	  instance = new Network(networkName);
	}
	
	public static Network getInstance() {
	  if (instance == null) instance = new Network("network");
	  return instance;
	}

  public static void createInstance(Network readObject) {
    instance = readObject;
    
  }
	
}
